#!/usr/bin/env python3
"""
Legacy setup.py for compatibility with tools like fpm.
Actual package configuration is in pyproject.toml.
"""
from setuptools import setup

if __name__ == "__main__":
    setup()
